#@version 6/1/2010
import data, copy, itertools, math, bussesoutput

class MainC:
    #---Private variables---
    __nodeDB = None

    #---Constructor---
    def __init__(self):
        self.__nodeDB = data.NodeDB()

    #---Database loaders---
    # Loads nodes from an existing NodeDB()
    def loadDB(self, nodeDB):
        self.__nodeDB = nodeDB
    # Loads nodes from a serialized file.
    def loadDBFromFile(self, nodeFileName = 'nodes.ser'):
        handler = data.DBHandler()
        handler.loadSerDB(nodeFileName)
        self.__nodeDB = handler.getDB()
    # Loads nodes from a text file.
    def loadDBFromText(self, nodeFileName = 'nodes.txt'):
        handler = data.DBHandler()
        handler.loadTextDB(nodeFileName)
        self.__nodeDB = handler.getDB()

    # Runs the program.
    # Print the solution in format [[route],cost,maxChildren,numBuses]
    def run(self, debug=False):
        solution = self.__findBestRoute('depot', 'school', debug)
        print 'Solution:'
        print 'Route: ' +str(solution[0])
        print 'Cost: ' +str(solution[1])
        print 'Total Children: ' +str(solution[2])
        print 'Number of Buses: ' +str(solution[3])
        print '--------------------------------------------'
        bussesoutput.read(solution)
        return solution        

    #---Algorithm stuff---
    # Finds all possible routes from one node to another. Once a previous max is reached, a new search
    # begins.
    def __findBestRoute(self, node1, node2, debug):
        maxChildren = 0
        for node in self.__nodeDB.getDB():
            maxChildren += self.__nodeDB.get(node).numChildren()
            
        routes = self.__findRoutes(copy.deepcopy(self.__nodeDB), node1, node2, [], 0, 0)

        bus = 1
        newRoutes = []
        bestRoute = [[],-1,0]
        while (routes != []):
            temp = self.__organizeRoutes(bus, routes, maxChildren)
            if len(temp) > 0:
                if bestRoute[1] is -1:
                    bestRoute = temp[0] + [bus]
                else:
                    for r in temp:
                        if r[1] < bestRoute[1]:
                            bestRoute = r + [bus]
                        
                # newRoutes is a list of all bus routes
                newRoutes += [bus] + temp
            bus += 1
            if (bus > (len(routes)/4)): #there are more buses than possible routes so loop is stopped
                break
        if debug:
            print newRoutes
        return bestRoute
        

    # Organizes all found routes to work with buses.
    def __organizeRoutes(self, valid, routes, maxC):
        
        iterables = itertools.combinations(routes,valid)
        modRoutes = []

        while (iterables):
            try:
                temp = iterables.next()
                r = [[],0,0]
                
                while len(temp) > 0:
                    r[0] += temp[0][0]
                    r[1] += temp[0][1]
                    r[2] += temp[0][2]
                    
                    if (r[2] > maxC):
                        r[1] = r[1] * math.ceil(r[2]/maxC)
                    
                    temp = temp[1:]
                modRoutes += [r]
            except StopIteration:
                break

        rMap = []
        for route in modRoutes:
            check = True
            
            for loc in self.__nodeDB.getDB():
                if not loc in route[0]:
                    check = False
                
            if check and (route[2] == maxC):
                rMap.append(route)
                if route in routes:
                    routes.remove(route)
        return rMap
    
    # Finds all possible routes with this node map.
    def __findRoutes(self, nodes, currNode, endNode, currRoute, currCost, currChildren):
        
        newChildren = currChildren + nodes.get(currNode).numChildren()
        nodes.get(currNode).deactivate()
        currRoute += [currNode]

        if (currNode != endNode):
            neighborDict = nodes.get(currNode).getNeighbors()
            routes = []
            deadEnd = True

            for neighbor in neighborDict:
                if nodes.get(neighbor).isActive():
                    deadEnd = False
                
            if not deadEnd:
                for neighbor in neighborDict:
                    if nodes.get(neighbor).isActive():

                        newCost = currCost + neighborDict[neighbor]
                        temp = self.__findRoutes(copy.deepcopy(nodes), neighbor, endNode, copy.copy(currRoute), newCost, newChildren)
                        
                        if temp is not None:
                            routes += temp
                        
                return routes
        else:
            return [[currRoute, currCost, newChildren]]
